import os
import requests
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from sklearn.preprocessing import StandardScaler
from model_training import update_lstm_model, train_lstm_model
from settings import settings
import json

data_dir = os.path.join(os.path.dirname(__file__), '..', 'data')

def get_stock_codes_from_csv(file_path):
    """Read stock codes from the CSV file."""
    try:
        if not os.path.exists(file_path):
            raise FileNotFoundError(f"CSV file not found: {file_path}")
        
        df = pd.read_csv(file_path, dtype={'code': str}, usecols=['code'])
        
        if df.empty:
            raise ValueError("CSV file is empty.")
        
        return df['code'].tolist()
        
    except Exception as e:
        print(f"Error reading stock codes from CSV file: {e}")
        return []

def get_real_time_data(ticker):
    url = f"{settings['trading_url']}/uapi/domestic-stock/v1/quotations/inquire-price"
    headers = {
        "Content-Type": "application/json; charset=utf-8",
        "Authorization": f"Bearer {get_valid_token()}",
        "appkey": settings['appkey'],
        "appsecret": settings['appsecret'],
        "tr_id": "FHKST01010100"
    }
    params = {
        "FID_COND_MRKT_DIV_CODE": "J",
        "FID_INPUT_ISCD": ticker
    }
    try:
        response = requests.get(url, headers=headers, params=params)
        response.raise_for_status()
        data = response.json()
        if data['rt_cd'] == '0':
            return pd.DataFrame([data['output']])
        else:
            print(f"Error fetching data for {ticker}: {data['msg1']}")
            return None
    except Exception as e:
        print(f"Error fetching data for {ticker}: {e}")
        return None

def get_valid_token():
    token_path = settings['token_path']
    try:
        with open(token_path, 'r') as file:
            token_data = json.load(file)
            if is_token_valid(token_data):
                return token_data['access_token']
            else:
                return refresh_token()
    except (FileNotFoundError, json.JSONDecodeError):
        return refresh_token()

def is_token_valid(token_data):
    expiration = datetime.strptime(token_data['expiration'], '%Y-%m-%d %H:%M:%S')
    return expiration > datetime.now()

def refresh_token():
    url = f"{settings['trading_url']}/oauth2/tokenP"
    data = {
        'grant_type': 'client_credentials',
        'appkey': settings['appkey'],
        'appsecret': settings['appsecret']
    }
    headers = {'Content-Type': 'application/json'}
    try:
        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            response_data = response.json()
            token_data = {
                'access_token': response_data['access_token'],
                'expiration': (datetime.now() + timedelta(seconds=response_data['expires_in'])).strftime('%Y-%m-%d %H:%M:%S')
            }
            with open(settings['token_path'], 'w') as file:
                json.dump(token_data, file)
            return token_data['access_token']
        else:
            raise Exception(f"Failed to refresh token: {response.text}")
    except requests.exceptions.RequestException as e:
        raise Exception(f"Failed to refresh token: {e}")
    
def normalize_data(df):
    """Normalize the data using StandardScaler."""
    scaler = StandardScaler()
    normalized_df = pd.DataFrame(scaler.fit_transform(df), columns=df.columns)
    return normalized_df, scaler    

def create_dataset(dataset, time_step=1):
    X, Y = [], []
    for i in range(len(dataset) - time_step - 1):
        a = dataset[i:(i + time_step), 0]
        X.append(a)
        Y.append(dataset[i + time_step, 0])
    return np.array(X), np.array(Y)

def real_time_prediction(model, scaler, ticker, time_step=60):
    real_time_data = get_real_time_data(ticker)
    if real_time_data is None:
        return None
    
    real_time_data = real_time_data[['stck_prpr']].rename(columns={'stck_prpr': 'close'})
    
    # 데이터 길이 확인
    if len(real_time_data) < time_step:
        print(f"Not enough data for the given time_step={time_step}. Data length={len(real_time_data)}")
        return None
    
    real_time_data, _ = normalize_data(real_time_data)
    scaled_real_time_data = real_time_data['close'].values.reshape(-1, 1)
    
    X_real_time, _ = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        print("Not enough data points after creating dataset.")
        return None
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], 1)
    
    predictions = model.predict(X_real_time)
    predictions = scaler.inverse_transform(predictions)
    return predictions

def update_real_time_model(model, scaler, ticker, time_step=60):
    real_time_data = get_real_time_data(ticker)
    if real_time_data is None:
        return model
    
    real_time_data = real_time_data[['stck_prpr']].rename(columns={'stck_prpr': 'close'})
    
    # 데이터 길이 확인
    if len(real_time_data) < time_step:
        print(f"Not enough data for the given time_step={time_step}. Data length={len(real_time_data)}")
        return model
    
    real_time_data, _ = normalize_data(real_time_data)
    scaled_real_time_data = real_time_data['close'].values.reshape(-1, 1)
    
    X_real_time, y_real_time = create_dataset(scaled_real_time_data, time_step)
    if len(X_real_time) == 0:
        print("Not enough data points after creating dataset.")
        return model
    
    X_real_time = X_real_time.reshape(X_real_time.shape[0], X_real_time.shape[1], 1)
    
    model = update_lstm_model(pd.DataFrame(y_real_time), model, time_step)
    return model

if __name__ == '__main__':
    from data_preprocessing import preprocess_data
    from model_training import train_lstm_model
    from config import *

    stock_codes_path = os.path.join(data_dir, 'stock_codes.csv')
    stock_codes = get_stock_codes_from_csv(stock_codes_path)

    try:
        for STOCK_TICKER in stock_codes:
            stock_data = get_real_time_data(STOCK_TICKER)
            if stock_data is not None:
                financial_data = pd.DataFrame()  # Placeholder
                news_data = pd.DataFrame()  # Placeholder
                economic_data = pd.DataFrame()  # Placeholder

                (X_train, X_test, y_train, y_test), scaler, combined_data = preprocess_data(stock_data, financial_data, news_data, economic_data)

                model = train_lstm_model(X_train, y_train)

                real_time_predictions = real_time_prediction(model, scaler, STOCK_TICKER)
                if real_time_predictions is not None:
                    print(f"Predictions for {STOCK_TICKER}: {real_time_predictions}")
                else:
                    print(f"Failed to fetch real-time predictions for {STOCK_TICKER}.")
    except Exception as e:
        print(f"An error occurred: {e}")
